/*
 * Project:  Motion Simulator V1
 * File:  main.c
 * Name:  Alex Pippin
 * 
 * Description:
 *  This program uses 3 ADC inputs to control the position of the platform held by 4 strings in the corners 
 *  connected to 1 DC toy motor per string. The ADC inputs are used to control the x, y, and z destination
 *  coordinates in comparison to the current calculated position of the platform.
 *
 * Inputs:
 *  PORTF.0   ->  ADC   -> x position
 *  PORTF.1   ->  ADC   -> y position
 *  PORTF.2   ->  ADC   -> z position
 *  PORTE.4   ->  INT4  -> Pushbutton -> Update position
 * 
 * Outputs:
 *  OC1A  ->  PORTB.5  PIN11  ->  Back Left   ->  FORWARD
 *  OC1B  ->  PORTB.6  PIN12  ->  Back Left   ->  REVERSE                       Y
 *  OC3A  ->  PORTE.3  PIN 5  ->  Back Right  ->  FORWARD       |BL         BR|
 *  OC3B  ->  PORTE.4  PIN 2  ->  Back Right  ->  REVERSE     Z |     Top     |
 *  OC4A  ->  PORTH.3  PIN 6  ->  Front Left  ->  FORWARD       |     View    |
 *  OC4B  ->  PORTH.4  PIN 7  ->  Front Left  ->  REVERSE       |FL         FR|
 *  OC3C  ->  PORTE.5  PIN 3  ->  Front Right ->  FORWARD               X
 *  OC4C  ->  PORTH.5  PIN 8  ->  Front Right ->  REVERSE
 *  PORTA.4:7 ->  
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <math.h>

// This is the TOP value for the PWMs to set the frequency to 4kHz
#define PWM_TOP 4209
// Move time in milliseconds
#define MOVE_TIME 10000
// Cross Diameter in inches
#define LINEAR_MODIFIER 1.6
// Theoretical Max positions
#define MAX_POS 100
// Motor Speed Slope based on PWM Duty Percentage and RPM
#define SLOPE (((64.52-28.16)/(11-5))+0.35)
// PWM Voltage logic
#define PWM_LOGIC_HIGH 5
// This is the conversion ratio between the virtual grid system and the  of the base in inches
#define CONV 22/100
// Diameter of the spool in inches
#define DIAMETER 0.25

#define BL_MOTOR 1
#define BR_MOTOR 2
#define FL_MOTOR 3
#define FR_MOTOR 4

#define BL_MOTOR_SPD_MOD 1
#define BR_MOTOR_SPD_MOD 1
#define FL_MOTOR_SPD_MOD 1
#define FR_MOTOR_SPD_MOD 1

volatile uint8_t update_pos = 0;

// Function Prototypes
void init_ports();
void set_speed(uint8_t, float, uint8_t);
float vect_2D(int16_t, int16_t);
float vect_3D(int16_t, int16_t, int16_t);
float duty_cycle_func(float, uint16_t);
void init_PWM1();
void init_PWM3();
void init_PWM4();
void init_ADC();
uint16_t read_ADC(uint16_t);
void ext_int0_init();

int main(void)
{
  init_ports();
  init_ADC();
  init_PWM1();
  init_PWM3();
  init_PWM4();
  ext_int0_init();

  // Variable Initialization
  uint8_t x = 50;
  uint8_t y = 50;
  uint8_t z = 50;
  uint8_t new_x = 0;
  uint8_t new_y = 0;
  uint8_t new_z = 0;
  
  while (1) {
    if (update_pos) {
      update_pos = 0;
      
      new_x = read_ADC(0) * 50UL / 512; // These are out of 100
      new_y = read_ADC(1) * 50UL / 512;
      new_z = read_ADC(2) * 50UL / 512;
      
      // Calculate the duty cycle for each individual motor
      float duty_bl = duty_cycle_func((vect_3D(x, MAX_POS - z, MAX_POS - y) - vect_3D(new_x, MAX_POS - new_z, MAX_POS - new_y)), MOVE_TIME);
      float duty_br = duty_cycle_func((vect_3D(MAX_POS - x, MAX_POS - z, MAX_POS - y) - vect_3D(MAX_POS - new_x, MAX_POS - new_z, MAX_POS - new_y)), MOVE_TIME);
      float duty_fl = duty_cycle_func((vect_3D(x, z, MAX_POS - y) - vect_3D(new_x, new_z, MAX_POS - new_y)), MOVE_TIME);
      float duty_fr = duty_cycle_func((vect_3D(MAX_POS - x, z, MAX_POS - y) - vect_3D(MAX_POS - new_x, new_z, MAX_POS - new_y)), MOVE_TIME);
      
      // Set the PWM value to each motor
      set_speed(BL_MOTOR, duty_bl, ((vect_2D(x, MAX_POS - z) < vect_2D(new_x, MAX_POS - new_z)) ? 1 : 0));
      set_speed(BR_MOTOR, duty_br, ((vect_2D(MAX_POS - x, MAX_POS - z) < vect_2D(MAX_POS - new_x, MAX_POS - new_z)) ? 1 : 0));
      set_speed(FL_MOTOR, duty_fl, ((vect_2D(x, z) < vect_2D(new_x, new_z)) ? 1 : 0));
      set_speed(FR_MOTOR, duty_fr, ((vect_2D(MAX_POS - x, z) < vect_2D(MAX_POS - new_x, new_z)) ? 1 : 0));
      
      delay(MOVE_TIME); // Delay for desired amount of time
      
      // Reset all PWM values
      OCR1A = 0;
      OCR1B = 0;
      OCR3A = 0;
      OCR3B = 0;
      OCR3C = 0;
      OCR4A = 0;
      OCR4B = 0;
      OCR1C = 0;
      
      // Update current position
      x = new_x;
      y = new_y;
      z = new_z;
    }
  }
}

/*
 * Initialize PORTs
 */
void init_ports() {
  DDRA = 0x03;  // PORTA.0:1 Inputs
  PORTA = ~0x03;  // Enable Pullup Resistors on input pins
  
  DDRF = 0x07;  // PORTF.0:2 Inputs
  PORTF = 0xFF; // Disable Pullup Resistors on PORTF.0:2
  
  DDRB = ~0x70; // PORTB.4:6 Outputs
  PORTB = 0x70; // Disable Pullup Resistors on output pins
  
  DDRH = ~0x40; // PORTH.6 Output
  PORTH = 0x40; // Disable pullup resistors on output pin
}

/*
 * Set the speed of designated motor to the given duty cycle
 * and given direction.
 */
void set_speed(uint8_t motor_sel, float duty, uint8_t dir) {
  switch (motor_sel) {
    case BL_MOTOR:
      if (dir) {  // Forward
        OCR1A = (uint16_t)(duty * (PWM_TOP + 1));
        OCR1B = 0;
      } else {    // Backward
        OCR1A = 0;
        OCR1B = (uint16_t)(duty * (PWM_TOP + 1));
      }
      break;
    case BR_MOTOR:
      if (dir) {  // Forward
        OCR3A = (uint16_t)(duty * (PWM_TOP + 1));
        OCR3B = 0;
      } else {    // Backward
        OCR3A = 0;
        OCR3B = (uint16_t)(duty * (PWM_TOP + 1));
      }
      break;
    case FL_MOTOR:
      if (dir) {  // Forward
        OCR4A = (uint16_t)(duty * (PWM_TOP + 1));
        OCR4B = 0;
      } else {    // Backward
        OCR4A = 0;
        OCR4B = (uint16_t)(duty * (PWM_TOP + 1));
      }
      break;
    case FR_MOTOR:
      if (dir) {  // Forward
        OCR3C = (uint16_t)(duty * (PWM_TOP + 1));
        OCR1C = 0;
      } else {    // Backward
        OCR3C = 0;
        OCR1C = (uint16_t)(duty * (PWM_TOP + 1));
      }
      break;  
  }
  return;
}

/*
 * Calculate the hypotenuse of a 2D triangle.
 */
float vect_2D(int16_t x, int16_t y) {
  return sqrt(pow(x, 2) + pow(y, 2));
}

/*
 * Calculate the hypotenuse of a 3D triangle.
 */
float vect_3D(int16_t l, int16_t w, int16_t h) {
  return sqrt(pow(l, 2) + pow(w, 2) + pow(h, 2));
}

/*
 * Calculate the needed duty cycle to produce the desired rotation speed of the motor shaft.
 */
float duty_cycle_func(float dist, uint16_t time_ms) {
  float duty = (dist * CONV * 60 * SLOPE) / (time_ms * M_PI * DIAMETER * 12);
  if (duty < 0) { duty = duty * -1; }
  return duty;
}

/*
 * Initialize ADC
 */
void init_ADC() {
  ADCSRA = (1<<ADEN) | (1<<ADPS1) | (1<<ADPS0);
}

/*
 * Read the 10-bit value from the ADC register.
 */
uint16_t read_ADC(uint16_t channel) {
  ADMUX = (ADMUX & 0xE0) | channel;   // Select channel
  ADCSRA |= (1<<ADSC);  // Start conversion
  
  while((ADCSRA & (1<<ADSC)) == 1); // Wait to finish conversion
  
  uint16_t result = ADCL;
  result |= (ADCH<<8);
  
  return result;
}

/*
 * Initialize PWM1's output
 */
 void init_PWM1(void) { //On digital 11
  DDRB = 0b11100000; //set the OC1A, B, & C pin as output
  TCCR1A = (1<<WGM11) | (1<<COM1A1) | (1<<COM1B1) | (1<<COM1C1);
  TCCR1B = (1<<WGM13) | (1<<WGM11) | (1<<CS11); //Set to pwm mode 14 with prescaler 8
  ICR1 = 5000;  // Set Frequency
  OCR1A = 2500; // Set Default PWM Duty Cycle to 50%
}

/*
 * Initialize PWM3's Output
 */
 void init_PWM3(void) { //On digital 5
  DDRE = 0b00111000; //set the OC3A, B, & C pin as output
  TCCR3A = (1<<WGM31) | (1<<COM3A1) | (1<<COM3B1) | (1<<COM3C1);
  TCCR3B = (1<<WGM33) | (1<<WGM31) | (1<<CS31); //Set to pwm mode 14 with prescaler 8
  ICR3 = 5000;  // Set Frequency
  OCR3A = 2500; // Set Default PWM Duty Cycle to 50%
}

/*
 * Initialize PWM4's Output
 */
 void init_PWM4(void) { //On digital 11
  DDRH = 0b00111000; //set the OCR4A, B, & C pin as output
  TCCR4A = (1<<WGM41) | (1<<COM4A1) | (1<<COM4B1);
  TCCR4B = (1<<WGM43) | (1<<CS41);  //Set to pwm mode 14 with prescaler 8
  ICR4 = 5000;  // Set Freq
  OCR4A = 2500; // Set Default PWM Duty Cycle to 50%
}

/*
 * Initialize External Interrupt 0
 */
void ext_int0_init() {
  EIMSK |= (1<<INT0);   // Enable interrupt pin on PD3
  EICRA |= (1<<ISC01) | (1<<ISC00);   // Configure for rising edge
}

ISR(INT0_vect) {
  update_pos = 1;
}
